from onepassword_tools.lib.OnePasswordUtils import OnePasswordUtils
from onepassword_local_search.OnePassword import OnePassword
from onepassword_tools.lib.MiscUtils import generate_password
from onepassword_tools.lib.OnePasswordServerItem import OnePasswordServerItem
from onepassword_tools.lib.ClickUtils import ClickUtils
import click
import copy
import sys


@click.command()
@click.option('--host', help='Host where the account is created', prompt=True, required=True)
@click.option('--username', help='Account username', prompt=True, required=True)
@click.option('--password', help='Password to use, default autogenerated')
@click.option('--password-length', help='Autogenerated password length, default 25', default=25, type=int)
@click.option('--vault', required=False, help='Vault uuid where to store the information')
def new_server_account(host, username, password, password_length, vault):
    """Create a new Server item in 1Password with the given credentials."""
    NewServerAccount(**locals()).run()


class NewServerAccount:

    entryTitleTemplate = 'USER %s ON %s'
    host: str = None
    onePassword: OnePassword
    onePasswordUtils: OnePasswordUtils
    password: str = None
    password_length: int = 25
    title: str = None
    username: str = None
    vault: str = None

    def __init__(self, **kwargs):
        self.onePasswordUtils = OnePasswordUtils()
        self.onePassword = OnePassword()
        self._init(**kwargs)
        if self.password is None:
            self.password = generate_password(self.password_length)
        self.title = self.entryTitleTemplate % (self.username, self.host)

    def _init(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def run(self):

        if not self.onePasswordUtils.is_authenticated():
            self.onePasswordUtils.authenticate()

        uuid = self.save_on_1password()
        if uuid:
            ClickUtils.success('Server password for %s was successfully saved in 1Password with uuid: %s'
                               % (self.host, uuid)
                               )
            sys.exit(0)
        else:
            ClickUtils.error('Unable to save entry in 1Password')
            sys.exit(1)

    def save_on_1password(self):
        arguments = vars(copy.copy(self))
        del arguments['onePassword']
        del arguments['onePasswordUtils']
        server_item = OnePasswordServerItem(**arguments)
        request_object = server_item.get_request_object()
        return self.onePasswordUtils.create_item(
            request_object=request_object,
            template=server_item.item_type,
            title=server_item.get_title(),
            vault=self.vault
        )



from onepassword_tools.lib.OnePasswordUtils import OnePasswordUtils
from onepassword_tools.lib.OnePasswordSSHKeyItem import OnePasswordSSHKeyItem
from onepassword_local_search.OnePassword import OnePassword
from onepassword_tools.lib.MiscUtils import generate_password
from onepassword_tools.lib.ClickUtils import ClickUtils
from onepassword_tools.lib.ConfigFile import ConfigFile
from getpass import getuser
from socket import gethostname
from onepassword_tools.lib.Crypto import generate_ssh_key
import click
import sys


def get_defaults():
    _defaults = {}
    config_file = ConfigFile()
    if '.' in __name__:
        command_name = __name__.rsplit('.', 1)[1]
    else:
        command_name = __name__
    command_config = config_file.get_section(command_name)

    # Grab from-host default value from config file or fallback to hostname
    if config_file.keys_exists(command_config, 'defaults', 'from-host'):
        _defaults['from-host'] = command_config['defaults']['from-host']
    else:
        _defaults['from-host'] = gethostname().encode('ascii', 'ignore').decode('ascii')

    # Grab from-user default value from config or fallback to current user name
    if config_file.keys_exists(command_config, 'defaults', 'from-user'):
        _defaults['from-user'] = command_config['defaults']['from-user']
    else:
        _defaults['from-user'] = getuser()
    return _defaults


defaults = get_defaults()


@click.command()
@click.option('--from-host', help='Host from where the SSH connection starts, default current hostname.', prompt=True, default=defaults['from-host'])
@click.option('--from-user', help='User who is responsible from initiating the connection, default current user.', prompt=True, default=defaults['from-user'])
@click.option('--to-host', help='Remote server hostname, prompted if empty.', prompt=True, required=True)
@click.option('--to-user', help='Remote user, prompted if empty.', prompt=True, required=True)
@click.option('--to-host-abbreviated', help='Alias of the remote server hostname used to initialize connection.', required=False)
@click.option('--no-passphrase', help='Create ssh key without passphrase.', prompt=False, required=False, default=False)
@click.option('--passphrase', help='Use this passphrase instead of an autogenerated one.')
@click.option('--passphrase-length', help='Length of the autogenerated passphrase.', default=50, type=int)
@click.option('--port', help='Remote port', required=False, default=22, type=int)
@click.option('--vault', help='1Password vault uuid where to store the SSH key', required=False, default=None)
def new_ssh_key(from_user, from_host, to_user, to_host, passphrase, passphrase_length, vault, port,
                to_host_abbreviated, no_passphrase):
    """Generates a new SSH key and store it in 1Password. Additional information are stored also to generate SSH config
    file when imported."""
    NewSSHKey(**locals()).run()


class NewSSHKey:

    entryTitleTemplate = 'id_rsa.from__%s@%s__to__%s@%s'
    from_user: str = None
    from_host: str = None
    no_passphrase: bool = None
    onePassword: OnePassword
    onePasswordUtils: OnePasswordUtils
    passphrase: str = None
    passphrase_length: int = None
    port: str = None
    tags = ['Clef SSH']
    title: str = None
    to_user: str = None
    to_host: str = None
    to_host_abbreviated: str = None
    vault: str = None

    def __init__(self, **kwargs):
        self._init(**kwargs)
        self.onePasswordUtils = OnePasswordUtils()
        self.onePassword = OnePassword()
        if not self.no_passphrase and not self.passphrase:
            self.passphrase = generate_password(self.passphrase_length)
        self.title = self.get_title(self.from_user, self.from_host, self.to_user, self.to_host)

    def __setitem__(self, key, value):
        if hasattr(self, key):
            setattr(self, key, value)

    def _init(self, **kwargs):
        for key, value in kwargs.items():
            self[key] = value

    @staticmethod
    def get_title(from_user, from_host, to_user, to_host):
        return NewSSHKey.entryTitleTemplate % (from_user, from_host, to_user, to_host)

    def run(self):
        if not self.onePasswordUtils.is_authenticated():
            self.onePasswordUtils.authenticate()

        public_key, private_key = generate_ssh_key(passphrase=self.passphrase)
        public_key += " %s" % self.title

        uuid = self.save_on_1password(public_key=public_key, private_key=private_key)

        if uuid:
            ClickUtils.success(
                'SSH Key %s was successfully saved in 1Password with uuid: %s' % (self.title, uuid))
            sys.exit(0)
        else:
            ClickUtils.error('Unable to save SSH Key in 1Password')
            sys.exit(1)

    def save_on_1password(self, public_key, private_key):
        arguments = vars(self)
        arguments['public_key'] = public_key
        arguments['private_key'] = private_key
        ssh_key_item = OnePasswordSSHKeyItem(**arguments)
        request_object = ssh_key_item.get_request_object()
        title = ssh_key_item.get_title()
        return self.onePasswordUtils.create_item(
            request_object=request_object,
            tags=self.tags,
            template=ssh_key_item.item_type,
            title=title,
            vault=self.vault
        )



from onepassword_tools.lib.ClickUtils import ClickUtils
from onepassword_tools.lib.ConfigFile import ConfigFile
from onepassword_tools.lib.Log import Log
from onepassword_tools.lib.MiscUtils import is_uuid
import json
import os
import secrets
import string
import subprocess
import sys
if os.environ.get('USE_LOCAL'):
    sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)) + "/../../../onepassword-local-search")
from onepassword_local_search.models.Item import Item
from onepassword_local_search.OnePassword import OnePassword
from onepassword_local_search.exceptions.ManagedException import ManagedException
import re


class OnePasswordUtils:

    onePassword: OnePassword
    suggestions: [] = []

    def __init__(self):
        self.onePassword = OnePassword()
        self.config = ConfigFile()

    @staticmethod
    def _communicate(subproc, inputstr=None):
        """
        Encode the input given to the subprocess if any
        :param subproc:
        :param inputstr:
        :return: A tuple of (stdout, stderr)
        """
        if inputstr is None:
            return subproc.communicate()
        if not inputstr:
            inputstr = ''
        return subproc.communicate(inputstr.encode())

    def _authenticate(self, shorthand=''):
        """
        Authenticate over 1Password and register the session key in the environment variable
        :return: Nothing
        """
        try:
            sessionKey = subprocess.check_output(['op', 'signin', shorthand, '--output=raw']) \
                .decode('utf-8') \
                .replace('\n', '')
            os.environ['OP_SESSION_' + shorthand] = sessionKey
        except subprocess.CalledProcessError:
            ClickUtils.error('Failed to authenticate. You may have written the wrong password ?')
            sys.exit(1)

    def authenticate(self):
        accounts = self.config.get_section('accounts')
        if len(accounts) > 0:
            for account in accounts:
                self._authenticate(account)
        else:
            self._authenticate(self.onePassword.configFileService.get_latest_signin())

    def create_item(self, request_object, template, title, tags=None, url='', vault=''):
        if tags is None:
            tags = []
        try:
            Log.debug(request_object, 'request data')
            rc, output, error = self.op_cli('encode', json.dumps(request_object))
            if output is not None and len(output) > 1:
                encrypted_data = output[:-1]
            else:
                raise Exception("Error while encoding json")

            command = 'create item "%s" %s' % (template, encrypted_data)
            if title and title != '':
                command += ' --title="%s"' % title
            if type(tags).__name__ != 'list':
                tags = [tags]
            if tags and len(tags) > 0:
                command += ' --tags="%s"' % ','.join(tags)
            if url and url != '' and template == 'Login':
                command += ' --url="%s"' % url
            if vault and vault != '':
                command += ' --vault="%s"' % vault

            Log.debug(command, 'op command executed')
            rc, output, error = self.op_cli(command)

            if output is not None and len(output) > 1:
                created_item = json.loads(output.replace('\n', ''))
                if not created_item.get('uuid'):
                    Log.error(error)
                    raise Exception("Error while creating onepassword entry")
                else:
                    return created_item.get('uuid')
            else:
                Log.error(error)
                raise Exception("Error while creating onepassword entry")

        except Exception:
            raise Exception("Entry while creating onepassword entry")

    def is_authenticated(self, check_mode=None) -> bool:
        """
        Check if we are authenticated over 1Password by decrypting local keys
        and trying to grab a template remotely
        :return: bool
        """
        is_authenticated_local = self.onePassword.is_authenticated()
        if check_mode is None:
            is_authenticated_remote = self.op_cli('get template Login')[0] == 0
            if is_authenticated_local and is_authenticated_remote:
                return True
        elif check_mode == 'local':
            if is_authenticated_local:
                return True
        elif check_mode == 'remote':
            is_authenticated_remote = self.op_cli('get template Login')[0] == 0
            if is_authenticated_remote:
                return True
        ClickUtils.error('You are not authenticated.')
        return False

    def generate_op_field_uuid(self):
        return self.generate_op_uuid(26)

    def generate_op_section_uuid(self):
        return self.generate_op_uuid(29)

    @staticmethod
    def generate_op_uuid(length=26):
        return ''.join(secrets.choice(string.ascii_lowercase) for _ in range(length))

    def get_alias(self, search):
        """
        Search if the input string is aliases to an uuid in the config file
        :param search:
        :return:
        """
        if self.config.config_key_exists('aliases', search):
            return self.config.config['aliases'][search]
        else:
            return search

    def op_cli(self, cmd, inputstr=None):
        """
        Call the 1Password cli (op)
        :param cmd: the op arguments and options
        :param inputstr: the input given to the command
        :return: A tuple of (returncode, stdout, stderr)
        """
        e = os.environ.copy()
        p = subprocess.Popen('op ' + cmd,
                             bufsize=0,
                             close_fds=True,
                             env=e,
                             shell=True,
                             stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE
                             )
        try:
            out, err = self._communicate(p, inputstr=inputstr)
        except subprocess.CalledProcessError:
            out = ''
            err = ''
        rc = p.returncode
        return rc, out.decode('utf-8'), err.decode('utf-8')

    def search_item_uuid_by_title(self, search):
        """
        Search an item by title
        :param search:
        :return: Item if only one match, None either
        """
        items = self.onePassword.get_items(search)
        if len(items) == 1:
            return self.onePassword.get(items[0].uuid, output=False)
        else:
            self.suggestions += items
            return None

    def try_to_grab_item(self, search) -> Item:
        """
        First try to grab by using 1Password or custom UUID, then fallback to
        search item by title.
        :param search:
        :return: Item or None
        """
        search = self.get_alias(search)
        try:
            if is_uuid(search):
                item = OnePassword(custom_uuid_mapping="UUID").get(search, output=False)
            elif re.match('[0-9]+]', search):
                item = OnePassword(custom_uuid_mapping="LASTPASS").get(search, output=False)
            else:
                item = self.onePassword.get(search, output=False)
        except ManagedException:
            item = self.search_item_uuid_by_title(search)
        return item



from onepassword_tools.lib.Log import Log
from onepassword_tools.lib.OnePasswordUtils import OnePasswordUtils
from onepassword_tools.lib.ClickUtils import ClickUtils
from onepassword_local_search.OnePassword import OnePassword
from onepassword_local_search.models.Item import Item
from plumbum import local
from string import Template
import click
import pexpect
import sys
import os
import textwrap


@click.command()
@click.argument('search', required=False, default=None)
@click.option('-D', help='cleanup ssh agent and remove all 1Password managed keys and configuration', is_flag=True)
@click.option('--no-ssh-config', help='Do not create ssh config file', default=False)
def ssh_add(search, d, no_ssh_config):
    """Loads a SSH key stored in 1Password by searching [SEARCH] in uuid or in item title, and creates a ssh
    configuration file of the following format:

    Match originalhost <Hostname> user <Remote user>\n
      IdentitiesOnly yes\n
      IdentityFile <path to key file>\n
      Hostname <Hostname>\n
      User <Remote user>\n
      Port <Port>
    """
    if d:
        SSHAdd().cleanup()
    else:
        if not search:
            Log.error('Error: Missing argument "SEARCH".')
            sys.exit(1)
        SSHAdd(no_ssh_config=no_ssh_config).add(search)


sshadd = local['ssh-add']


class SSHAdd:

    item: Item
    keyFilePath: str
    keyStoragePath: str
    no_ssh_config: bool
    onePassword: OnePassword
    onePasswordUtils: OnePasswordUtils
    passPhrase: str
    privateKey: {}
    sessionKey: str
    sshConfigFilePath: str
    sshConfigStoragePath: str
    suggestions: [] = []
    title: str

    def __init__(self, no_ssh_config=False):
        self.onePasswordUtils = OnePasswordUtils()
        self.onePassword = OnePassword()
        self.keyStoragePath = os.path.expandvars(os.path.join('$HOME', '.ssh', 'keys.1password.d'))
        self.sshConfigStoragePath = os.path.expandvars(os.path.join('$HOME', '.ssh', 'config.1password.d'))
        self.init_path(self.keyStoragePath)
        self.init_path(self.sshConfigStoragePath)
        self.no_ssh_config = no_ssh_config

    def _get_ssh_config_match_original_host(self):
        hosts = [ self.item.get('Hostname', strict=False) ]
        abbr = self.item.get('Hostname abbreviated', strict=False)
        if abbr:
            hosts.append(abbr)
        return 'originalhost %s' % ','.join(hosts)

    def _get_ssh_config_match_user(self):
        to_user = self.item.get('Remote user', strict=False)
        if to_user:
            if os.environ.get('USER') == to_user:
                return ''
            else:
                return 'user %s' % to_user
        return ''

    def add(self, search):

        if not self.onePasswordUtils.is_authenticated(check_mode='local'):
            self.onePasswordUtils.authenticate()

        item = self.onePasswordUtils.try_to_grab_item(search)
        if item is None:
            if len(self.suggestions) > 0:
                ClickUtils.error('Unable to find proper match. Please refine your search')
                #TODO: use percol to select entry
                ClickUtils.info('Matching entries:')
                for suggestion in self.suggestions:
                    ClickUtils.info('%s %s' % (suggestion.uuid, suggestion.overview['title']))
            elif len(self.suggestions) == 0:
                ClickUtils.error('Unable to find any items matching your search')
            sys.exit(1)

        self.title = item.get('title')

        if self.key_has_been_registered(self.title):
            ClickUtils.success('Key %s already registered.' % self.title)
            sys.exit(0)

        self.item = item
        self.privateKey = self.get_private_key(item)
        self.passPhrase = self.get_pass_phrase(item)
        self.register_ssh_key(self.privateKey, self.passPhrase)

        if not self.no_ssh_config:
            self.create_ssh_config_file()

        if self.key_has_been_registered(self.title):
            ClickUtils.success('Key %s has been added to agent successfully.' % self.title)
            sys.exit(0)
        else:
            if os.path.isfile(self.keyFilePath):
                os.unlink(self.keyFilePath)
            if os.path.isfile(self.sshConfigStoragePath):
                os.unlink(self.sshConfigStoragePath)
            ClickUtils.error('Unable to add key %s to agent' % self.title)
            sys.exit(1)

    def cleanup(self):

        for file in os.scandir(self.sshConfigStoragePath):
            if file.name.endswith('.cfg'):
                os.unlink(file.path)

        for file in os.scandir(self.keyStoragePath):
            if file.name.endswith('.key'):
                os.unlink(file.path)

        sshadd['-D']()
        ClickUtils.success('All dynamic keys and SSH configurations have been removed.')

    def create_ssh_config_file(self):

        config = textwrap.dedent(Template("""\
            Match $original_host $user
              IdentitiesOnly yes
              IdentityFile $private_key_file_path
              Hostname $to_host""").substitute(
            original_host=self._get_ssh_config_match_original_host(),
            private_key_file_path=self.keyFilePath,
            to_host=self.item.get('Hostname', strict=False),
            user=self._get_ssh_config_match_user()
        ))

        to_user = self.item.get('Remote user', strict=False)
        if to_user:
            config += '\n  User %s' % to_user

        port = self.item.get('port', strict=False)
        if port:
            config += '\n  Port %s' % port

        ssh_config_file_path = os.path.join(self.sshConfigStoragePath, self.title.replace('id_rsa.', '') + '.cfg')
        with open(ssh_config_file_path, 'w+') \
                as config_file:
            config_file.write(config)
            config_file.flush()
            os.chmod(config_file.name, 0o600)
            self.sshConfigFilePath = ssh_config_file_path

    @staticmethod
    def get_private_key(key):
        private_key = key.get('Private Key', strict=False)
        if private_key:
            return private_key
        ClickUtils.error('unable to get private key')
        sys.exit(1)

    @staticmethod
    def get_pass_phrase(key):
        passphrase = key.get('Passphrase', strict=False)
        if passphrase:
            return passphrase
        ClickUtils.error('unable to get passphrase')
        sys.exit(1)

    @staticmethod
    def init_path(path):
        path = os.path.expandvars(path)
        if not os.path.isdir(path):
            os.mkdir(path, 0o700)
        elif (os.stat(path).st_mode & 0o777) != 0o700:
            os.chmod(path, 0o700)

    @staticmethod
    def key_has_been_registered(name):
        registered_keys = sshadd('-l', retcode=None)
        return name.replace('id_rsa.', '') in registered_keys

    def register_ssh_key(self, private_key, passphrase):

        with open(os.path.join(self.keyStoragePath, self.title + '.key'), 'w+') as key_file:
            key_file.write(private_key)
            key_file.flush()
            self.keyFilePath = key_file.name
            os.chmod(self.keyFilePath, 0o600)

        if passphrase:
            command = 'ssh-add %s' % self.keyFilePath
            bash = pexpect.spawn('bash', ['-c', command])
            bash.expect_exact('Enter passphrase')
            bash.sendline(passphrase)
            bash.expect_exact(pexpect.EOF)
        else:
            sshadd[self.keyFilePath]()




from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends.openssl.backend import backend as openssl_backend


def generate_ssh_key(passphrase=None, key_size=4096):
    key = rsa.generate_private_key(
        backend=openssl_backend,
        public_exponent=65537,
        key_size=key_size
    )

    encryption = crypto_serialization.BestAvailableEncryption(passphrase.encode('utf-8')) if passphrase \
        else crypto_serialization.NoEncryption()

    private_key = key.private_bytes(
        crypto_serialization.Encoding.PEM,
        crypto_serialization.PrivateFormat.PKCS8,
        encryption
    )

    public_key = key.public_key().public_bytes(
        crypto_serialization.Encoding.OpenSSH,
        crypto_serialization.PublicFormat.OpenSSH
    )

    return public_key.decode('utf-8'), private_key.decode('utf-8')